import json
import shlex
from abc import abstractmethod
from contextlib import contextmanager
from copy import deepcopy
from os import getenv, getcwd, chdir, environ
from os.path import join, basename, normpath, abspath
from typing import Optional, List, Generator, Dict, Tuple, Union

import click


from jsonschema.validators import Draft7Validator


__all__ = [
    # Contexts
    'GametaContext', 'gameta_context',
]


SHELL = getenv('SHELL', '/bin/sh')


class File(object):
    """
    Generic file interface for Gameta file formats

    Attributes:
        context (GametaContext): Reference to Gameta Context
        file_name (str): Name of the reference file
    """

    def __init__(self, context: 'GametaContext', file_name: str):
        self.context = context
        self.file_name = file_name

    @property
    def file(self) -> str:
        """
        Returns the absolute path to the reference file

        Returns:
            str: Absolute path to the file
        """
        return join(self.context.project_dir, self.file_name)

    @abstractmethod
    def load(self) -> None:
        """
        Abstractmethod to load data and validate data from the file and populate the GametaContext

        Returns:
            None
        """

    @abstractmethod
    def export(self) -> None:
        """
        Abstractmethod to export data from the GametaContext to the file

        Returns:
            None
        """


class GitIgnore(File):
    """
    Interface for the .gitignore file

    Attributes:
        context (GametaContext): Reference to Gameta Context
        file_name (str): Reference to the .gitignore file
    """
    def __init__(self, context: 'GametaContext', file_name: str = '.gitignore'):
        super(GitIgnore, self).__init__(context, file_name)

    def load(self) -> None:
        """
        Loads data from the .gitignore file and populates the GametaContext

        Returns:
            None
        """
        try:
            with open(self.file, 'r') as f:
                self.context.gitignore_data = f.readlines()
        except FileNotFoundError:
            return
        except Exception as e:
            self.context.gitignore_data = []
            click.echo(f"Could not load {self.file_name} file due to: {e.__class__.__name__}.{str(e)}")

    def export(self) -> None:
        """
        Exports data from the GametaContext to the .gitignore file

        Returns:
            None
        """
        try:
            with open(self.file, 'w') as f:
                f.writelines(self.context.gitignore_data)
        except Exception as e:
            click.echo(f"Could not export data to {self.file_name} file: {e.__class__.__name__}.{str(e)}")


class Meta(File):
    """
    Interface for the .meta file

    Attributes:
        context (GametaContext): Reference to Gameta Context
        file_name (str): Reference to the .meta file
    """

    def __init__(self, context: 'GametaContext', file_name: str = '.meta'):
        super(Meta, self).__init__(context, file_name)

    def load(self) -> None:
        """
        Loads data from the .meta file, validates it and populates the GametaContext

        Returns:
            None
        """
        # Attempt to load .meta file
        try:
            with open(self.file_name, 'r') as f:
                self.context.gameta_data = json.load(f)
        except FileNotFoundError:
            return
        except Exception as e:
            click.echo(f"Could not load {self.file_name} file due to: {e.__class__.__name__}.{str(e)}")

        # Validate repositories
        try:
            for repo in self.context.gameta_data['projects'].values():
                self.context.validators['repositories'].validate(repo)
            self.context.repositories = self.context.gameta_data['projects']
            self.context.is_metarepo = True
            self.context.generate_tags()
        except Exception as e:
            self.context.repositories = {}
            self.context.tags = {}
            click.echo(f"Malformed repository element, error: {e.__class__.__name__}.{str(e)}")

        # Validate commands
        try:
            for command in self.context.gameta_data.get('commands', {}).values():
                self.context.validators['commands'].validate(command)
            self.context.commands = self.context.gameta_data.get('commands', {})
        except Exception as e:
            self.context.commands = {}
            click.echo(f"Malformed commands element, error: {e.__class__.__name__}.{str(e)}")

        # Validate constants
        try:
            self.context.validators['constants'].validate(self.context.gameta_data.get('constants', {}))
            self.context.constants = self.context.gameta_data.get('constants', {})
        except Exception as e:
            self.context.constants = {}
            click.echo(f"Malformed constants element, error: {e.__class__.__name__}.{str(e)}")

    def export(self) -> None:
        """
        Exports data from the GametaContext to the .meta file

        Returns:
            None
        """
        try:
            self.context.gameta_data['projects'] = self.context.repositories
            if self.context.commands:
                self.context.gameta_data['commands'] = self.context.commands
            if self.context.constants:
                self.context.gameta_data['constants'] = self.context.constants
            with open(self.file, 'w') as f:
                json.dump(self.context.gameta_data, f, indent=2)
        except Exception as e:
            click.echo(f"Could not export data to {self.file_name} file: {e.__class__.__name__}.{str(e)}")


class GametaContext(object):
    """
    GametaContext for the current Gameta session

    Attributes:
        __schema__ (Dict): JSON Schema for Gameta .meta file
        validators (Dict[str, jsonschema.Draft7Validator]): JSON Schema validators for each object component
        reserved_params (Dict[str, List[str]): Reserved parameters for each object group

        project_dir (Optional[str]): Project directory
        is_metarepo (bool): Project is a metarepo
        gameta_data (Dict): Gameta data extracted and exported
        repositories (Dict[str, Dict]): Data of all the repositories contained in the metarepo
        tags (Dict[str, List[str]]): Repository data organised according to tags
        constants (Dict[str, Union[str, int, bool, float]]): Gameta constants data extracted
        commands (Dict): Gameta commands data extracted
        gitignore_data (List[str]): Gitignore data extracted from the .gitignore file
        env_vars (Dict): Extracted environment variables with keys prefixed with $
        files (Dict[str, File]): File formats supported
    """
    __schema__: Dict = {
        '$schema': "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "properties": {
            "repositories": {
                "$ref": "#/definitions/repositories"
            },
            "commands": {
                "$ref": "#/definitions/commands"
            },
            "constants": {
                "$ref": "#/definitions/constants"
            },
            "required": [
                "repositories"
            ]
        },
        'definitions': {
            "repositories": {
                "type": "object",
                "properties": {
                    "url": {
                        "type": ["string", "null"],
                        "format": "uri"
                    },
                    "path": {
                        "type": "string"
                    },
                    "tags": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "__metarepo__": {
                        "type": "boolean"
                    }
                },
                "required": [
                    "url", "path", "__metarepo__"
                ]
            },
            "commands": {
                "type": "object",
                "properties": {
                    "commands": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                    },
                    "raise_errors": {
                        "type": "boolean"
                    },
                    "shell": {
                        "type": "boolean"
                    },
                    "python": {
                        "type": "boolean"
                    },
                    "verbose": {
                        "type": "boolean"
                    },
                    "repositories": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                    },
                    "tags": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                    }
                },
                "minProperties": 6,
                "maxProperties": 7,
                "additionalProperties": False,
            },
            "constants": {
                "type": "object",
                "propertyNames": {
                    "pattern": "^[A-Z0-9_-]"
                }
            }
        }
    }

    validators = {
        'meta': Draft7Validator(__schema__),
        'repositories': Draft7Validator(__schema__['definitions']['repositories']),
        'commands': Draft7Validator(__schema__['definitions']['commands']),
        'constants': Draft7Validator(__schema__['definitions']['constants'])
    }

    reserved_params: Dict[str, List[str]] = {
        'repositories': list(__schema__['definitions']['repositories']['properties'].keys()),
        'commands': list(__schema__['definitions']['commands']['properties'].keys())
    }

    def __init__(self):
        self.project_dir: Optional[str] = None
        self.gitignore_data: List[str] = []
        self.is_metarepo: bool = False
        self.gameta_data: Dict = {}
        self.constants: Dict[str, Union[str, int, bool, float]] = {}
        self.commands: Dict = {}
        self.repositories: Dict[str, Dict] = {}
        self.tags: Dict[str, List[str]] = {}

        self.env_vars: Dict = {
            '$' + k.upper(): v
            for k, v in environ.items()
        }

        self.files: Dict[str, File] = {
            'meta': Meta(self),
            'gitignore': GitIgnore(self)
        }

    @property
    def project_name(self) -> str:
        """
        Returns the name of the project

        Returns:
            str: Name of the project
        """
        return basename(self.project_dir)

    @property
    def meta(self) -> str:
        """
        Returns the path to the .meta file of the project, i.e. where it should be if the Project has not been
        initialised

        Returns:
            str: Path to the project's .meta file
        """
        return self.files['meta'].file

    @property
    def gitignore(self) -> str:
        """
        Returns the path to the .gitignore file of the project, i.e. where it should be if the Project has not been
        initialised

        Returns:
            str: Path to the project's .gitignore file
        """
        return self.files['gitignore'].file

    def add_gitignore(self, path: str) -> None:
        """
        Adds the path to the gitignore_data

        Args:
            path (str): Path to be added

        Returns:
            None
        """
        self.gitignore_data.append(path + '/\n')

    def remove_gitignore(self, path: str) -> None:
        """
        Removes the path from the gitignore_data

        Args:
            path (str): Path to be removed

        Returns:
            None
        """
        try:
            self.gitignore_data.remove(path + '/\n')
        except ValueError:
            return

    def is_primary_metarepo(self, repo: str) -> bool:
        """
        Returns a boolean if the repository is a primary meta-repository

        Args:
            repo (str): Repository to check

        Returns:
            bool: Flag to indicate if repository is a primary meta-repository
        """
        return abspath(self.repositories[repo]["path"]) == self.project_dir

    def load(self) -> None:
        """
        Loads data from all supported file formats

        Returns:
            None
        """
        for file, interface in self.files.items():
            interface.load()

    def export(self) -> None:
        """
        Exports data to all supported file formats

        Returns:
            None
        """
        for file, interface in self.files.items():
            interface.export()

    def generate_tags(self) -> None:
        """
        Updates the tag indexes of the repositories

        Returns:
            None
        """
        for repo, details in self.repositories.items():
            for tag in details.get('tags', []):
                if tag in self.tags:
                    self.tags[tag].append(repo)
                else:
                    self.tags[tag] = [repo]

    def apply(
            self,
            commands: List[str],
            repos: List[str] = (),
            shell: bool = False,
            python: bool = False,
    ) -> Generator[Tuple[str, str], None, None]:
        """
        Yields a list of commands to all repositories or a selected set of them, substitutes relevant parameters stored
        in .meta file

        Args:
            commands (List[str]): Commands to be applied
            repos (List[str]): Selected set of repositories
            shell (bool): Flag to indicate if a separate shell should be used
            python (bool): Flag to indicate if commands are to be tokenised as Python commands

        Returns:
            None
        """
        repositories: List[Tuple[str, Dict[str, str]]] = \
            [(repo, details) for repo, details in self.repositories.items() if repo in repos] or \
            list(self.repositories.items())

        for repo, details in repositories:
            # Generate complete set of parameters for substitution
            combined_details: Dict = deepcopy(details)
            combined_details.update(self.constants)
            combined_details.update(self.env_vars)
            combined_details.update(
                {
                    '__repos__':
                        json.dumps(self.repositories)
                        .replace("true", "True")
                        .replace("false", "False")
                        .replace("null", "None")
                }
            )

            with self.cd(combined_details['path']):
                repo_commands: List[str] = [c.format(**combined_details) for c in deepcopy(commands)]
                if python:
                    command: List[str] = self.python(repo_commands)
                elif shell:
                    command: List[str] = self.shell(repo_commands)
                else:
                    command: List[str] = self.tokenise(' && '.join(repo_commands))
                yield repo, command

    @staticmethod
    def tokenise(command: str) -> List[str]:
        """
        Tokenises the commands into a form that is readily acceptable by subprocess

        Args:
            command (str): Constructed commands to be tokenised

        Returns:
            List[str]: Tokenised commands
        """
        return shlex.split(command)

    @contextmanager
    def cd(self, sub_directory: str) -> Generator[str, None, None]:
        """
        Changes directory to a subdirectory within the project

        Args:
            sub_directory (str): Relative subdirectory within the project

        Returns:
            Generator[str, None, None]: Path to current directory
        """
        cwd = getcwd()
        path = normpath(join(self.project_dir, sub_directory.lstrip('/')))
        chdir(path)
        yield path
        chdir(cwd)

    def shell(self, commands: List[str]) -> List[str]:
        """
        Prepares commands to be executed in a separate shell as subprocess does not natively handle piping

        Args:
            commands (List[str]): User-defined commands

        Returns:
            List[str]: Shell command string to be executed by subprocess
        """
        return self.tokenise(
            f'{SHELL} -c "' +
            ' && '.join(commands) +
            '"'
        )

    def python(self, commands: List[str]) -> List[str]:
        """
        Prepares commands to be executed by Python interpreter via shell

        Args:
            commands List[str]: Python scripts

        Returns:
            List[str]: Python prepared commands to be executed by subprocess
        """
        return self.shell(
            ["python3 -c \'{}\'".format(command.replace('"', '\\\"')) for command in commands]
        )


gameta_context = click.make_pass_decorator(GametaContext, ensure=True)




from os import getcwd, chdir
from os.path import abspath

import click

from .context import gameta_context, GametaContext
from gameta import __version__


__all__ = [
    # Gameta CLI
    'gameta_cli'
]


@click.group('gameta', invoke_without_command=True)
@click.option('--project-dir', '-d', type=str, default=getcwd())
@click.option('--version', '-v', is_flag=True, default=False)
@gameta_context
def gameta_cli(context: GametaContext, project_dir: str, version: bool) -> None:
    """
    Gameta CLI, contains commands to simplify your DevOps operations
    \f
    Args:
        context (GametaContext): Gameta Context
        project_dir (str): Project directory, defaults to the current working directory
        version (bool): Flag to print Gameta's version and exit

    Returns:
        None

    Examples:
        $ gameta -v  # Prints version and exits
        $ gameta -d /path/to/project/dir  # Specify a project directory
    """
    if version:
        click.echo(f"Gameta version: {__version__}")
        return

    context.project_dir = abspath(project_dir)
    chdir(context.project_dir)

    # Load current repositories
    context.load()


from .init import *
from .repos import *
from .tags import *
from .apply import *
from .params import *
from .cmd import *
from .constants import *



from copy import deepcopy
from typing import Tuple, Dict, Optional, Callable, Any

import click

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['command_cli']


@gameta_cli.group('cmd')
@gameta_context
def command_cli(context: GametaContext) -> None:
    """
    CLI for managing CLI commands
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Raises:
        click.ClickException: If we are not currently operating in a metarepo directory
    """
    if not context.is_metarepo:
        raise click.ClickException(f"{context.project_dir} is not a metarepo, initialise it with 'gameta init'")


@command_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Gameta command name to be added to the store')
@click.option('--overwrite', '-o', type=bool, is_flag=True, default=False,
              help='Overwrite existing Gameta command in the store')
@click.option('--command', '-c', 'commands', type=str, required=True, multiple=True, help='CLI Commands to be executed')
@click.option('--tags', '-t', type=str, multiple=True, default=(), help='Repository tags to apply CLI commands to')
@click.option('--repositories', '-r', type=str, multiple=True, default=(), help='Repositories to apply CLI commands to')
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Display execution output when CLI command is applied')
@click.option('--shell', '-s', is_flag=True, default=False, help='Execute CLI commands in a separate shell')
@click.option('--python', '-p', is_flag=True, default=False, help='Execute Python scripts using Python 3 interpreter')
@click.option('--raise-errors', '-e', is_flag=True, default=False,
              help='Raise errors that occur when executing CLI commands and terminate execution')
@gameta_context
def add(
        context: GametaContext,
        name: str,
        overwrite: bool,
        commands: Tuple[str],
        tags: Tuple[str],
        repositories: Tuple[str],
        verbose: bool,
        shell: bool,
        python: bool,
        raise_errors: bool
) -> None:
    """
    Adds a new Gameta command to the Gameta command store
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the CLI command to be stored
        overwrite (bool): Flag to overwrite the existing Gameta command if it exists, defaults to false

        commands (Tuple[str]): CLI command to be applied
        tags (Tuple[str]): Repository tags to apply command to
        repositories (Tuple[str]): Repositories to apply command to
        verbose (bool): Flag to indicate that output should be displayed as the command is applied
        shell (bool): Flag to indicate that command should be executed in a separate shell
        python (bool): Flag to indicate that command should be executed by the Python 3 interpreter
        raise_errors (bool): Flag to indicate that errors should be raised if they occur during execution and the
                             overall execution should be terminated

    Returns:
        None

    Raises:
        click.ClickException: If errors occur during processing
    """
    # Validate parameters
    # Tags must already be added before they can be used
    if any(t not in context.tags for t in tags):
        raise click.ClickException(
            f"One of the tags in {list(tags)} has not been added, please run `gameta tags add` to add it first"
        )
    # Repositories must already be added before they can be used
    if any(r not in context.repositories for r in repositories):
        raise click.ClickException(
            f"One of the repositories in {list(repositories)} does not exist, please run `gameta repo add` to "
            f"add it first"
        )
    # If Python flag is set, all commands must be valid Python scripts
    if python:
        try:
            for command in commands:
                compile(command, 'test', 'exec')
        except SyntaxError:
            raise click.ClickException(f"One of the commands in {list(commands)} is not a valid Python script")

    try:
        gameta_command: Dict = {
            'commands': list(commands),
            'tags': list(tags),
            'repositories': list(repositories),
            'verbose': verbose,
            'shell': True if len(commands) > 1 else shell,
            'python': python,
            'raise_errors': raise_errors
        }

        click.echo(f"Adding command {name} with parameters ({gameta_command}) to the command store")
        if name in context.commands:
            if overwrite is True:
                click.echo(f"Overwriting command {name} in the command store")
            else:
                click.echo(f"Command {name} already exists in the command store")
                return

        context.commands[name] = gameta_command
        context.export()
        click.echo(f"Successfully added command {name} to the command store")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@command_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Name of the Gameta command to be deleted')
@gameta_context
def delete(context: GametaContext, name: str) -> None:
    """
    Deletes an existing Gameta command from the Gameta command store
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the Gameta command to be deleted

    Returns:
        None

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Deleting command {name} from the command store")
    if name not in context.commands:
        raise click.ClickException(f"Command {name} does not exist in the command store")

    try:
        del context.commands[name]
        context.export()
        click.echo(f"Successfully deleted command {name} from the command store")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@command_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Gameta command name to be added to the store')
@click.option('--command', '-c', 'commands', type=str, default=None, multiple=True,
              help='New CLI commands to be executed')
@click.option('--tags', '-t', type=str, multiple=True, default=None,
              help='New repository tags to apply CLI commands to')
@click.option('--repositories', '-r', type=str, multiple=True, default=None,
              help='New repositories to apply CLI commands to')
@click.option('--verbose/--no-verbose', '-v/-nv', is_flag=True, default=None,
              help='Display execution output when CLI command is applied')
@click.option('--shell/--no-shell', '-s/-ns', is_flag=True, default=None,
              help='Execute CLI commands in a separate shell')
@click.option('--python/--no-python', '-p/-np', is_flag=True, default=None,
              help='Execute Python scripts using Python 3 interpreter')
@click.option('--raise-errors/--no-errors', '-e/-ne', is_flag=True, default=None,
              help='Raise errors that occur when executing CLI commands and terminate execution')
@gameta_context
def update(
        context: GametaContext,
        name: str,
        commands: Optional[Tuple[str]],
        tags: Optional[Tuple[str]],
        repositories: Optional[Tuple[str]],
        verbose: Optional[bool],
        shell: Optional[bool],
        python: Optional[bool],
        raise_errors: Optional[bool]
) -> None:
    """
    Updates an existing Gameta command in the Gameta command store
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Command to be updated
        commands (Tuple[str]): CLI command to be applied
        tags (Tuple[str]): Repository tags to apply command to
        repositories (Tuple[str]): Repositories to apply command to
        verbose (Optional[bool]): Flag to indicate that output should be displayed as the command is applied
        shell (Optional[bool]): Flag to indicate that command should be executed in a separate shell
        python (Optional[bool]): Flag to indicate that command should be executed by the Python 3 interpreter
        raise_errors (Optional[bool]): Flag to indicate that errors should be raised if they occur during execution and
                                       the overall execution should be terminated

    Returns:
        None

    Raises:
        click.ClickException: If errors occur during processing
    """

    click.echo(f"Updating command {name} in the command store")
    updates: Dict = {
        'commands': commands,
        'tags': tags,
        'repositories': repositories,
        'verbose': verbose,
        'shell': shell,
        'python': python,
        'raise_errors': raise_errors
    }
    if name not in context.commands:
        raise click.ClickException(f"Command {name} does not exist in the command store")

    try:
        updated_command: Dict = deepcopy(context.commands[name])
        for key, value in updates.items():
            # Skip invalid data
            if not (
                isinstance(value, bool) or
                (isinstance(value, tuple) and len(value) > 0)
            ):
                continue

            updated_command[key] = value

        if 'python' not in updated_command:
            updated_command['python'] = False

        # Validate parameters
        # Multiple commands need to be accompanied with the shell parameter
        if len(updated_command['commands']) > 1 and updated_command['shell'] is False:
            raise click.ClickException('Multiple CLI commands requires shell param to be True')
        # Tags must already be added before they can be used
        if any(t not in context.tags for t in tags):
            raise click.ClickException(
                f"One of the tags in {list(tags)} has not been added, please run `gameta tags add` to add it first"
            )
        # Repositories must already be added before they can be used
        if any(r not in context.repositories for r in repositories):
            raise click.ClickException(
                f"One of the repositories in {list(repositories)} does not exist, please run `gameta repo add` to "
                f"add it first"
            )
        # If Python flag is set, all commands need to be valid Python scripts
        if updated_command['python']:
            try:
                for command in updated_command['commands']:
                    compile(command, 'test', 'exec')
            except SyntaxError:
                raise click.ClickException(f"One of the commands in {list(commands)} is not a valid Python script")
        # If Python flag was unset, all the commands cannot be valid Python scripts
        if context.commands[name].get('python', False) is True and updated_command['python'] is False:
            for command in updated_command['commands']:
                try:
                    compile(command, 'test', 'exec')
                    raise click.ClickException(
                        f"Python flag was unset but one of the commands in {updated_command['commands']} "
                        f"is still Python compilable"
                    )
                except SyntaxError:
                    continue

        context.commands[name] = updated_command
        context.export()
        click.echo(f"Successfully updated command {name} in the command store")
    except click.ClickException:
        raise
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@command_cli.command()
@gameta_context
def ls(context: GametaContext) -> None:
    """
    Lists all commands in the Gameta command store
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Raises:
        click.ClickException: If errors occur during processing
    """
    def format_command(name: str, details: Dict) -> str:
        """
        Formats a command to be printed as a string

        Args:
            name (str): Command name
            details (Dict): Command details to be printed

        Returns:
            str: Formatted command string
        """
        command_string: str = f'{name}:\n'
        param_string: str = '{}: {}\n'
        formatters: Dict[str, Callable[[Any], str]] = {
            'commands': lambda v: ' && '.join(v),
            'tags': lambda v: ', '.join(v),
            'repositories': lambda v: ', '.join(v)
        }
        for key in ['commands', 'tags', 'repositories', 'verbose', 'shell', 'python', 'raise_errors']:
            if key in details:
                command_string += '\t' + param_string.format(key, formatters.get(key, str)(details.get(key)))

        return command_string

    try:
        for command, command_details in context.commands.items():
            click.echo(format_command(command, command_details))
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@command_cli.command()
@click.option('--command', '-c', 'commands', type=str, multiple=True, required=True,
              help='Gameta commands to be invoked')
@click.pass_context
def exec(context: click.Context, commands: Tuple[str]) -> None:
    """
    Executes Gameta commands from the CLI command store
    \f
    Args:
        context (click.Context): Click Context
        commands (Tuple[str]): Gameta commands to be executed

    Returns:
        None

    Raises:
        click.ClickException: If errors occur during processing
    """
    from gameta.apply import apply

    g_context: GametaContext = context.obj
    if any(c not in g_context.commands for c in commands):
        raise click.ClickException(
            f"One of the commands in {list(commands)} does not exist in the Gameta command store, please run "
            f"`gameta cmd add` to add it first"
        )

    try:
        click.echo(f"Executing {list(commands)}")
        for command in commands:
            click.echo(f"Executing Gameta command {command}")
            context.invoke(apply, **g_context.commands[command])
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")



import subprocess
from typing import List, Tuple

import click

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['apply']


@gameta_cli.command()
@click.option('--command', '-c', 'commands', type=str, required=True, multiple=True, help='CLI Commands to be executed')
@click.option('--tags', '-t', type=str, multiple=True, default=(), help='Repository tags to apply CLI commands to')
@click.option('--repositories', '-r', type=str, multiple=True, default=(), help='Repositories to apply CLI commands to')
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Display execution output when CLI commands are applied')
@click.option('--shell', '-s', is_flag=True, default=False, help='Execute CLI commands in a separate shell')
@click.option('--python', '-p', is_flag=True, default=False, help='Execute Python scripts using Python 3 interpreter')
@click.option('--raise-errors', '-e', is_flag=True, default=False,
              help='Raise errors that occur when executing CLI commands and terminate execution')
@gameta_context
def apply(
        context: GametaContext,
        commands: Tuple[str],
        tags: Tuple[str],
        repositories: Tuple[str],
        verbose: bool,
        shell: bool,
        python: bool,
        raise_errors: bool
) -> None:
    """
    Applies a CLI command to all repositories (by default) or a specific set of repositories
    \f
    Args:
        context (GametaContext): Gameta Context
        commands (Tuple[str]): CLI command to be applied
        tags (Tuple[str]): Repository tags to apply command to
        repositories (Tuple[str]): Repositories to apply command to
        verbose (bool): Flag to indicate that output should be displayed as the command is applied
        python (bool): Flag to indicate that command should be executed with the Python 3 interpreter
        shell (bool): Flag to indicate that command should be executed in a separate shell
        raise_errors (bool): Flag to indicate that errors should be raised if they occur during execution and the
                             overall execution should be terminated

    Returns:
        None

    Examples:
        $ gameta apply -c "git fetch --all --tags --prune" -c "git checkout {branch}"  # Multiple commands
        $ gameta apply -c "git fetch --all --tags --prune" -t tag1 -t tag2 -t tag3 -r repo_a  # Apply to tags and repos
        $ gameta apply -c "git fetch --all --tags --prune" -e  # Raise errors and terminate
        $ gameta apply -c "git fetch --all --tags --prune" -s  # Executed in a separate shell
        $ gameta apply -c "git fetch --all --tags --prune" -v  # Verbose

    Raises:
        click.ClickException: If errors occur during processing
    """
    try:
        repos: List[str] = sorted(
            list(
                set([repo for tag in tags for repo in context.tags.get(tag, [])]) |
                set([repo for repo in repositories if repo in context.repositories])
            )
        )

        if python:
            try:
                for command in commands:
                    compile(command, 'test', 'exec')
            except SyntaxError:
                raise click.ClickException(f"One of the commands in {list(commands)} is not a valid Python script")

        # Python subprocess does not handle multiple commands
        # hence we need to handle it in a separate shell
        if len(commands) > 1:
            click.echo("Multiple commands detected, executing in a separate shell")
            shell = True

        if shell:
            click.echo(
                f"Applying {list(commands)} to repos {repos if repos else list(context.repositories.keys())} "
                f"in a separate shell"
            )
        elif python:
            click.echo(
                f"Applying Python commands {list(commands)} to repos "
                f"{repos if repos else list(context.repositories.keys())} in a separate shell"
            )
        else:
            click.echo(
                f"Applying {list(commands)} to repos {repos if repos else list(context.repositories.keys())}"
            )
        for repo, c in context.apply(list(commands), repos=repos, shell=shell, python=python):
            click.echo(f"Executing {' '.join(c)} in {repo}")
            try:
                if verbose:
                    click.echo(subprocess.check_output(c))
                else:
                    subprocess.check_output(c)
            except subprocess.SubprocessError as e:
                if raise_errors:
                    raise click.ClickException(
                        f'Error {e.__class__.__name__}.{str(e)} occurred when executing command {" ".join(c)} in {repo}'
                    )
                click.echo(
                    f'Error {e.__class__.__name__}.{str(e)} occurred when executing command {" ".join(c)} in {repo}, '
                    f'continuing execution'
                )
    except click.ClickException:
        raise
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")



from typing import List, Dict, Tuple

import click

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['tags_cli']


@gameta_cli.group('tags')
@gameta_context
def tags_cli(context: GametaContext) -> None:
    """
    CLI for adding tags to repos
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Raises:
        click.ClickException: If we are not currently operating in a metarepo directory
    """
    if not context.is_metarepo:
        raise click.ClickException(f"{context.project_dir} is not a metarepo, initialise it with 'gameta init'")


@tags_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Name of the repository to add tags to')
@click.option('--tags', '-t', type=str, required=True, multiple=True, help='Tags to add to the repository')
@gameta_context
def add(context: GametaContext, name: str, tags: List[str]) -> None:
    """
    Adds a set of tags to a repository
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the repository to add tags to
        tags (List[str]): Tags to add to the repository

    Returns:
        None

    Examples:
        $ gameta tags add -n repo_name -t tag1 -t tag2 -t tag3

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Adding tags {list(tags)} to {name}")
    if name not in context.repositories:
        raise click.ClickException(f"Repository {name} does not exist in .meta file")

    try:
        repo_details: Dict = context.repositories[name]
        repo_details['tags'] = sorted(list(set(repo_details.get('tags', [])) | set(tags)))
        context.export()
        click.echo(f"Successfully added tags to repository {name}")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@tags_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Name of the repository to delete tags from')
@click.option('--tags', '-t', type=str, required=True, multiple=True, help='Tags to delete from the repository')
@gameta_context
def delete(context: GametaContext, name: str, tags: Tuple[str]) -> None:
    """
    Deletes a set of tags from a repository
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the repository to delete tags from
        tags (Tuple[str]): Tags to delete from the repository

    Returns:
        None

    Examples:
        $ gameta tags delete -n repo_name -t tag1 -t tag2 -t tag3

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Deleting tags {list(tags)} from {name}")
    if name not in context.repositories:
        raise click.ClickException(f"Repository {name} does not exist in .meta file")

    tags: List[str] = list(tags)
    if context.is_primary_metarepo(name) and 'metarepo' in tags:
        click.echo("Unable to delete the metarepo tag from metarepo, removing it before deleting other tags")
        tags.remove('metarepo')

    try:
        repo_details: Dict = context.repositories[name]
        repo_details['tags'] = sorted(list(set(repo_details.get('tags', [])) - set(tags)))
        context.export()
        click.echo(f"Successfully deleted tags from repository {name}")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")



from copy import deepcopy
from os.path import join, normpath, exists
from shutil import rmtree, copytree
from typing import Dict, Optional

import click
from git import Repo, GitError

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['repo_cli']


@gameta_cli.group('repo')
@gameta_context
def repo_cli(context: GametaContext) -> None:
    """
    CLI for managing repos in metarepos
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Raises:
        click.ClickException: If we are not currently operating in a metarepo directory
    """
    if not context.is_metarepo:
        raise click.ClickException(f"{context.project_dir} is not a metarepo, initialise it with 'gameta init'")


@repo_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Name of the repository to be added')
@click.option('--url', '-u', type=str, required=True, help='URL of the git repository')
@click.option('--path', '-p', type=str, required=True, help='Relative path to clone the repository to')
@click.option('--overwrite', '-o', is_flag=True, default=False,
              help='Overwrite the existing repository configuration if it exists, defaults to false')
@gameta_context
def add(context: GametaContext, name: str, url: str, path: str, overwrite: bool) -> Optional[Dict]:
    """
    Adds a new repository to the metarepo
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the repository to be added
        url (str): URL of the git repository
        path (str): Local relative path to clone the repository to
        overwrite (bool): Flag to overwrite the existing repository configuration if it exists, defaults to false

    Returns:
        Optional[Dict]: Returns the repository details added

    Examples:
        $ gameta repo add -n repo_name -u https://github.com/git_user/repo_name.git -p path/to/repo
        $ gameta repo add -n repo_name -u https://github.com/git_user/repo_name.git -p path/to/repo -o

    Raises:
        click.ClickException: If errors occur during processing
    """
    clone_path: str = normpath(join(context.project_dir, path))
    click.echo(f"Adding git repository {name}, {url} to {clone_path}")
    try:
        if exists(clone_path):
            repo = Repo(clone_path)
            if repo.bare:
                raise click.ClickException(f'Path {clone_path} exists, please clear it before proceeding')
            if repo.remote().url != url:
                raise click.ClickException(
                    f'URL of repository at {clone_path} ({repo.remote().url}) does not match the requested url {url}'
                )
            click.echo(f"Repository {name} exists locally, skipping clone")
        else:
            Repo.clone_from(url, clone_path)
    except GitError as e:
        raise click.ClickException(f'Error cloning {name} into directory {clone_path}: {e.__class__.__name__}.{str(e)}')

    try:
        click.echo(f"Repository {name} has been added locally")
        if name in context.repositories:
            if overwrite is True:
                click.echo(f"Overwriting repository {name} in .meta file")
            else:
                click.echo(f"Repository {name} has already been added to .meta file")
                return
        else:
            click.echo(f"Adding {name} to .meta file")

        context.repositories[name] = {
            'url': url,
            'path': path,
            '__metarepo__': False
        }
        context.add_gitignore(path)
        context.export()

        click.echo(f"Successfully added repository {name}")
        return context.repositories[name]
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@repo_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Name of the repository to be deleted')
@click.option(' /--no-clear', ' /-c', 'clear', is_flag=True, default=True,
              help='Do not clear the repository folder locally')
@gameta_context
def delete(context: GametaContext, name: str, clear: bool) -> None:
    """
    Deletes the repository from the .meta file and locally if specified
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the repository to be deleted
        clear (bool): Flag to indicate if repository should be cleared locally, defaults to true

    Returns:
        None

    Examples:
        $ gameta repo delete -n repo_name
        $ gameta repo delete -n repo_name -c  # Does not clear local repository

    Raises:
        click.ClickException: If errors occur during processing
    """
    try:
        click.echo(f"Deleting repository {name} from .meta file")
        if name not in context.repositories:
            click.echo(f"Repository {name} does not exist in the .meta file, ignoring")
            return

        # Prevent user from deleting primary metarepo
        if context.is_primary_metarepo(name):
            raise click.ClickException(f"Cannot delete repository {name} as it is a metarepo")

        if clear:
            click.echo(f"Clearing repository {name} locally")
            rmtree(join(context.project_dir, context.repositories[name]["path"]), ignore_errors=True)

        context.remove_gitignore(context.repositories[name]["path"])
        del context.repositories[name]
        context.export()
        click.echo(f"Repository {name} successfully deleted")
    except click.ClickException:
        raise
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@repo_cli.command()
@click.option('--name', '-n', type=str, required=True, help='Name of the repository to be updated')
@click.option('--new-name', '-e', type=str, default=None, help='New repository name to update to')
@click.option('--new-url', '-u', type=str, default=None, help='New repository URL')
@click.option('--new-path', '-p', type=str, default=None,
              help='New relative path to clone the repository to')
@click.option(' /--no-sync', ' /-s', 'sync', is_flag=True, default=True, help='Does not sync all updates locally')
@gameta_context
def update(
        context: GametaContext,
        name: str,
        new_name: Optional[str],
        new_url: Optional[str],
        new_path: Optional[str],
        sync: Optional[bool],
) -> None:
    """
    Updates the key parameters defining a repository
    \f
    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the repository to be updated
        new_name (Optional[str]): New repository name
        new_url (Optional[str]): New repository URL
        new_path (Optional[str]): New relative path
        sync (bool): Sync all updates, defaults to True

    Returns:
        None

    Examples:
        $ gameta repo update -n orig_repo_name -e new_repo_name -u https://github.com/git_user/repo_name.git
                             -p /path/to/repo
        $ gameta repo update -n orig_repo_name -e new_repo_name -u https://github.com/git_user/repo_name.git
                             -p /path/to/repo -s  # Does not sync local changes

    Raises:
        click.ClickException: If errors occur during processing
    """
    if name not in context.repositories:
        raise click.ClickException(f"Repository {name} does not exist in the .meta file, please add it first")

    try:
        click.echo(f"Updating repository {name} with new details (name: {new_name}, url: {new_url}, path: {new_path})")
        curr_repo_details: Dict = context.repositories[name]

        # Update the details
        new_repo_details: Dict = deepcopy(context.repositories[name])
        for key, value in [('url', new_url), ('path', new_path)]:
            if value is not None:
                new_repo_details[key] = value

        if new_name is not None:
            del context.repositories[name]
            name = new_name

        # Perform a physical sync with the updated details
        if sync:
            click.echo("Performing a physical sync")
            if new_url is not None:
                # Only URL changes, need to clear the existing path and clone repository
                click.echo(f"Cloning repository from new URL: {new_url}")
                rmtree(join(context.project_dir, curr_repo_details['path']), ignore_errors=True)
                Repo.clone_from(new_url, normpath(join(context.project_dir, new_repo_details['path'])))
            elif new_url is None and new_path is not None:
                # Only path changes, need to move the directory to the new location
                click.echo(f"Copying repository to new path: {new_path}")
                copytree(
                    join(context.project_dir, curr_repo_details["path"]),
                    join(context.project_dir, new_repo_details["path"])
                )
                rmtree(join(context.project_dir, curr_repo_details['path']), ignore_errors=True)
            else:
                # No physical changes
                click.echo("No physical changes")
            click.echo("Sync complete")

        # Export the updated details
        context.remove_gitignore(curr_repo_details["path"])
        context.add_gitignore(new_repo_details["path"])
        context.repositories[name] = new_repo_details
        context.export()
        click.echo(f"Successfully updated repository {name} with new details")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@repo_cli.command()
@gameta_context
def ls(context: GametaContext) -> None:
    """
    Lists all the repositories added in the .meta file
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Examples:
        $ gameta repo ls

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Listing repositories managed in metarepo {context.project_dir}")
    for repo in context.repositories:
        click.echo(f"{repo}")



import builtins
from typing import Union

import click

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['constants_cli']


@gameta_cli.group('const')
@gameta_context
def constants_cli(context: GametaContext) -> None:
    """
    CLI for managing constants in metarepos
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Raises:
        click.ClickException: If we are not currently operating in a metarepo directory
    """
    if not context.is_metarepo:
        raise click.ClickException(f"{context.project_dir} is not a metarepo, initialise it with 'gameta init'")


@constants_cli.command()
@click.option("--name", '-n', type=str, required=True, help='Name of the constant to stored')
@click.option("--type", '-t', 'ctype', type=click.Choice(['int', 'float', 'str', 'bool']), required=True,
              help='Type of the constant to be stored as')
@click.option('--value', '-v', type=str, required=True, help='Value of the constant to be stored')
@gameta_context
def add(context: GametaContext, name: str, ctype: str, value: str) -> None:
    """
    Adds a set of constants to the constants store

    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the constant provided
        ctype (str): Type of the constant provided
        value (str): Value of the constant

    Returns:
        None

    Examples:
        $ gameta const add -n test -t str -v hello_world  # Adds constant TEST as a string to Gameta
        $ gameta const add -n test_1 -t float -v 1.0  # Adds constant TEST_1 as a float to Gameta

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Adding constant {name}")
    constant_type: type = getattr(builtins, ctype, str)

    try:
        name: str = name.upper()
        value: Union[int, float, bool, str] = click.types.convert_type(constant_type)(value)
        context.constants[name.upper()] = value
        context.export()
        click.echo(f"Successfully added constant {name}: {value} (type: {ctype}) to .meta file")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@constants_cli.command()
@click.option("--name", '-n', type=str, required=True, help="Name of constant to be deleted")
@gameta_context
def delete(context: GametaContext, name: str) -> None:
    """
    Deletes a Gameta constant from the constant store

    Args:
        context (GametaContext): Gameta Context
        name (str): Name of the constant to be deleted

    Returns:
        None

    Examples:
        $ gameta const delete -n test  # Deletes the TEST constant from the constant store

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Deleting constant {name}")
    try:
        if name in context.constants:
            del context.constants[name]
        elif name.upper() in context.constants:
            del context.constants[name.upper()]
        else:
            raise click.ClickException(f"Constant {name} does not exist in .meta file")

        context.export()
        click.echo(f"Successfully deleted constant {name} from .meta file")
    except click.ClickException:
        raise
    except Exception as e:
        raise click.ClickException(f'{e.__class__.__name__}.{str(e)}')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gameta']

package_data = \
{'': ['*']}

install_requires = \
['Click>=7.1.2,<8.0.0',
 'colorama>=0.4.3,<0.5.0',
 'gitpython>=3.1.9,<4.0.0',
 'jsonschema>=3.2.0,<4.0.0']

entry_points = \
{'console_scripts': ['gameta = gameta.cli:gameta_cli']}

setup_kwargs = {
    'name': 'gameta',
    'version': '0.2.2',
    'description': 'CLI Tool for managing meta-repositories',
    'long_description': '# Gameta\n\nGameta is a powerful CLI tool that helps you to manage \n[meta-repositories](https://genius-systems.github.io/gameta/metarepos) or \n[metarepos](https://genius-systems.github.io/gameta/metarepos). It allows \nyou to create links to related repositories, make changes and sync them, and \nprovides functionality to customise and apply other CLI commands to these \nrepositories.\n\n## What is Gameta?\n\nGameta is a play on the word gamete (reproductive cells), and similar \nto how gametes form the building blocks for life, gameta helps to \nmanage the many repositories that form the building blocks for more\ncomplex software.\n\n## System Requirements\n\nGameta requires Python 3.7+, it is designed to be multi-platform but currently is\nonly tested in Linux environments.\n\n## Installation\n\nGameta can be easily installed and updated via pip:\n\n```bash\n$ pip install gameta  # install\n$ pip install -U gameta  # update\n```\n\n## Getting Started\n\nGetting started is really easy.\n\n```bash\ngameta init\ngameta init -g  # To initialise directory as a Git repo \n```\n\nGameta will extract git information and create the .meta file which stores \nall your project configurations.\n\n```json\n{\n  "projects": {\n    "gameta": {\n      "path": ".",\n      "tags": ["metarepo"],\n      "url": "https://github.com:genius-systems/gameta.git"\n    }\n  }\n}\n```\n\n___\n**Note**\n\nIf your project has not been initialised with Git, you can specify the\n`--git` or `-g` flag to perform a git init, but you will have to update the\nrepository URL manually later on. \n___\n\nIf your repository contains a .meta file generated by Gameta, simply run the\nfollowing command to sync all linked repositories locally:\n\n```bash\ngameta sync\n```\n\n## Adding a Repository\n\nRun the following command to add a new repository.\n\n```bash\ngameta repo add -n GitPython -u https://github.com/gitpython-developers/GitPython.git -p GitPython\n```\n\n___\n**Note**\n\nThe path variable is the relative path within the metarepo itself\n___\n\nYou should see another entry under the projects object within the .meta file\n\n```json\n{\n  "projects": {\n    "gameta": {\n      "path": ".",\n      "tags": ["metarepo"],\n      "url": "git@github.com:genius-systems/gameta.git"\n    },\n    "GitPython": {\n      "path": "GitPython",\n      "tags": ["a", "b", "c"],\n      "url": "https://github.com/gitpython-developers/GitPython.git"\n    }\n  }\n}\n```\n\nYou should also see the repository cloned to the relative path specified\n\n## Applying commands\n\nGameta provides a powerful toolkit to manage your set of repositories.\n\n```bash\ngameta apply -c "git fetch --all --tags --prune" -c "git merge"\n```\n\nThe command above applies the following to all repositories:\n\n1. Fetches all git updates, tags and prunes redundant git artifacts\n2. Merges changes on the default branch\n\nThere\'s alot more that you can do with commands, see the \n[Applying Commands](https://genius-systems.github.io/gameta/user_guide/applying_commands)\npage for more information\n\n## Best practices\n\nLearn how to leverage on Gameta\'s capabilities in the \n[Best Practices](https://genius-systems.github.io/gameta/user_guide/best_practices) \npage\n\n## Documentation \n\nDocumentation can be found at [Gameta Docs](https://genius-systems.github.io/gameta)\n',
    'author': 'David Lee',
    'author_email': 'ltw_david@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://genius-systems.github.io/gameta',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from os.path import splitext, basename
from typing import Optional

import click

from git import Repo, InvalidGitRepositoryError, GitError

from .cli import gameta_cli
from .context import gameta_context, GametaContext


__all__ = ['init', 'sync']


@gameta_cli.command()
@click.option('--git', '-g', is_flag=True, default=False, help='Flag to initialise the current directory as a git repo')
@click.option('--overwrite', '-o', is_flag=True, default=False, help='Flag to overwrite existing .meta file')
@gameta_context
def init(context: GametaContext, overwrite: bool, git: bool) -> None:
    """
    Initialises a repository as a metarepo
    \f
    Args:
        context (GametaContext): Gameta Context
        overwrite (bool): Flag to indicate if existing .meta file should be overwritten
        git (bool): Flag to indicate if we should initialise git in the current directory

    Returns:
        None

    Examples:
        $ gameta init  # With default values
        $ gameta init -g  # Initialises the current working directory as a git repository
        $ gameta init -o  # Overwrites the existing .meta file if it exists

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Initialising metarepo in {context.project_dir}")

    try:
        repo: Repo = Repo(context.project_dir)
        name: str = splitext(basename(repo.remote().url))[0]
        url: Optional[str] = repo.remote().url
    except InvalidGitRepositoryError:
        if git is False:
            raise click.ClickException(f"{context.project_dir} is not a valid git repo, initialise it with -g flag")
        click.echo(f"Initialising {context.project_dir} as a git repository")
        Repo.init(context.project_dir)
        name: str = basename(context.project_dir)
        url: Optional[str] = None
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")

    try:
        if context.is_metarepo and overwrite is False:
            click.echo(f"{context.project_dir} is a metarepo, ignoring")
            return

        context.repositories[name] = {
            'url': url,
            'path': '.',
            'tags': ['metarepo'],
            '__metarepo__': True
        }
        context.export()
        click.echo(f"Successfully initialised {name} as a metarepo")
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")


@gameta_cli.command()
@gameta_context
def sync(context: GametaContext) -> None:
    """
    Syncs all the repositories listed in the .meta file locally
    \f
    Args:
        context (GametaContext): Gameta Context

    Returns:
        None

    Examples:
        $ gameta sync

    Raises:
        click.ClickException: If errors occur during processing
    """
    click.echo(f"Syncing all child repositories in metarepo {context.project_dir}")

    try:
        for repo, details in context.repositories.items():

            # We assume the metarepo has already been cloned
            if context.is_primary_metarepo(repo):
                continue

            try:
                Repo.clone_from(details['url'], details['path'])
                click.echo(f'Successfully synced {repo} to {details["path"]}')
            except GitError as e:
                click.echo(f"An error occurred {str(e)}, skipping repo")
                continue
    except Exception as e:
        raise click.ClickException(f"{e.__class__.__name__}.{str(e)}")
